#####################################################################
#                                                                   #
# /user_devices/FastComtecMCS8A/blacs_workers.py                    #
#                                                                   #
# Jan 2025, Leolab Cavity                                           #
#                                                                   #
#                                                                   #
#####################################################################

from . import fastcomtec_api
from blacs.tab_base_classes import Worker
from labscript_utils.labconfig import LabConfig
import os
import labscript_utils.h5_lock
import h5py
import numpy as np
import time
from datetime import datetime

class MCS8AWorker(Worker):

    def init(self):
        self.base_dir = LabConfig().get("DEFAULT","experiment_shot_storage") + "\\MCS8A"
        os.makedirs(self.base_dir, exist_ok=True) # create MCS8A folder if it doesn't exist yet

        self.device = fastcomtec_api.FastComTec()
        # print("Status: ", self.device.get_acq_status())
        if getattr(self.device.get_acq_status(),"started") == 1:
            time.sleep(1)
        self.device.halt_measurement()
        self.dataindex=0

        # load (default) config
        if not os.path.realpath(self.config_path) == self.config_path.replace("/","\\"):
            self.config_path = os.path.dirname(os.path.realpath(__file__)) + "\\" + self.config_path
            self.config_path = self.config_path.replace("\\", "/")
        print("Load config...", self.config_path)
        self.device.load_config(self.config_path)

        
        # Save data only in list file and in binary format
        self.device._send_dll_command("mpafmt=dat")
        self.device._send_dll_command("savedata=2")

        # Set how long each measurement lasts, i.e. how ofen a new file is started.
        # We don't want it too long, because then the single files get huge, but we also don't
        # want it too often, because each stopping and starting takes some time.
        self.maxduration = 600 # seconds
        self.device._send_dll_command("bitshift=18") # 400ps * 2^0x18 = 0.0067 s
        range_int = np.ceil(self.maxduration / (400e-12 * 2**0x18) / 64) * 64
        self.device._send_dll_command(f"range={int(range_int)}")
    
    def get_new_mpapath(self):
        mpaname = self.device_name + "_" + datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + ".mpa"
        return self.base_dir + "\\" + mpaname
    
    def transition_to_buffered(self, device_name, h5_file, initial_values, fresh):
        # If device is already started, we leave it running and just save the datapath and index in the h5file.
        # If the device is not running, or already longer than the defined duration, send the stop command, set new datafile and start new measurement.
        # TODO: How to handle parameters (start new measurment when they changed, or treat parameters statically)
        with h5py.File(h5_file,'r') as f: 
            stop_time = f["devices/" + f["connection table"].attrs["master_pseudoclock"]].attrs["stop_time"]
        if getattr(self.device.get_acq_status(),"started") != 1 or time.perf_counter()-getattr(self,"start_time",-self.maxduration)>self.maxduration-stop_time-3:
            if self.dataindex==0:
                self.device.halt_measurement()
            while getattr(self.device.get_acq_status(),"started") == 1:
                # If the current measurement doesn't have enought time left for this shot, wait until measurment ends. 
                time.sleep(0.1)
                print("Wait until last measurement stopped...", end="\r")
            self.device.halt_measurement()
            mpapath = self.get_new_mpapath()
            self.device._send_dll_command(f"mpaname={mpapath}")
            self.datapath = mpapath.replace(".mpa",".lst")
            self.dataindex = 0
            print("Set filename...", self.datapath.split("\\")[-1])

            # self.device.erase_measruement() # Don't erase because this takes super long for some reason
            self.device.start_measurement()
            self.start_time = time.perf_counter()
            time.sleep(2) # some buffer time to make sure the measurement already started for sure
            # TODO: What's the minimum time to wait here?

        # Save where to find the measurement data (which file and within file)
        with h5py.File(h5_file,'r+') as f:
            group = f[f"devices/{self.device_name}"]
            group.attrs["datapath"] = self.datapath
            group.attrs["dataindex"] = self.dataindex
            self.dataindex += group.attrs["n_trigger"]

        return {}

    def transition_to_manual(self):
        return True
    
    def program_manual(self, values):
        return {}
    
    def check_remote_values(self):
        return True

    def abort_buffered(self):
        return self.transition_to_manual()

    def abort_transition_to_buffered(self):
        return True
    
    def shutdown(self):
        self.device._send_dll_command("exit")